@@ -29,6 +29,7 @@
  */
 
 #include "common.h"
+#include "strgetopt.h"
 #include <pkcs11-helper-1.0/pkcs11h-token.h>
 #include <pkcs11-helper-1.0/pkcs11h-certificate.h>
 #include "command.h"
@@ -314,7 +315,7 @@ cleanup:
    S CERTINFO <certtype> <hexstring_with_id>
 
    If certificate is issuer, we set type to 102 (useful); otherwise it is
-   assumed that we're in posession of private key, so the type is set to 101
+   assumed that we're in possession of private key, so the type is set to 101
    (trusted).  The certificate ID is percent-plus escaped displayName.
 */
 static
@@ -516,6 +517,10 @@ send_certificate_list (
 			nameinfo = NULL;
 		}
 
+		if (error == GPG_ERR_WRONG_PUBKEY_ALGO) {
+			error = GPG_ERR_NO_ERROR;
+		}
+
 		if (error != GPG_ERR_NO_ERROR) {
 			goto cleanup;
 		}
@@ -528,7 +533,7 @@ cleanup:
 	return error;
 }
 
-int _get_certificate_by_name (assuan_context_t ctx, char *name, int typehint, pkcs11h_certificate_id_t *p_cert_id, char **p_key) {
+int _get_certificate_by_name (assuan_context_t ctx, const char *name, int typehint, pkcs11h_certificate_id_t *p_cert_id, const char **p_key) {
 	cmd_data_t *data = (cmd_data_t *)assuan_get_pointer (ctx);
 	gpg_err_code_t error = GPG_ERR_BAD_KEY;
 	pkcs11h_certificate_id_list_t user_certificates = NULL;
@@ -536,7 +541,7 @@ int _get_certificate_by_name (assuan_context_t ctx, char *name, int typehint, pk
 	pkcs11h_certificate_id_t cert_id = NULL;
 	char *key_hexgrip = NULL;
 	gcry_sexp_t sexp = NULL;
-	char *key = NULL;
+	const char *key = NULL;
 	int type;
 
 	*p_cert_id = NULL;
@@ -555,27 +560,32 @@ int _get_certificate_by_name (assuan_context_t ctx, char *name, int typehint, pk
 		type = typehint;
 	}
 	else if (strncmp (name, OPENPGP_KEY_NAME_PREFIX, strlen (OPENPGP_KEY_NAME_PREFIX))) {
-		return common_map_pkcs11_error (
+		if ((error = common_map_pkcs11_error (
 			pkcs11h_certificate_deserializeCertificateId (p_cert_id, name)
-		);
+		)) == GPG_ERR_NO_ERROR) {
+			goto cleanup;
+		}
+		key = name;
 	}
 	else {
 		type = atoi(name + strlen (OPENPGP_KEY_NAME_PREFIX));
 	}
 
-	switch (type) {
-		case OPENPGP_SIGN:
-			key = data->config->openpgp_sign;
-		break;
-		case OPENPGP_ENCR:
-			key = data->config->openpgp_encr;
-		break;
-		case OPENPGP_AUTH:
-			key = data->config->openpgp_auth;
-		break;
-		default:
-			error = GPG_ERR_BAD_KEY;
-			goto cleanup;
+	if (key == NULL) {
+		switch (type) {
+			case OPENPGP_SIGN:
+				key = data->config->openpgp_sign;
+			break;
+			case OPENPGP_ENCR:
+				key = data->config->openpgp_encr;
+			break;
+			case OPENPGP_AUTH:
+				key = data->config->openpgp_auth;
+			break;
+			default:
+				error = GPG_ERR_BAD_KEY;
+				goto cleanup;
+		}
 	}
 
 	if (key == NULL) {
@@ -685,6 +695,8 @@ gpg_error_t cmd_serialno (assuan_context_t ctx, char *line)
 	gpg_err_code_t error = GPG_ERR_GENERAL;
 	char *serial = NULL;
 
+	(void)line;
+
 	if (
 		(error = get_serial(ctx, &serial)) != GPG_ERR_NO_ERROR
 	) {
@@ -807,11 +819,14 @@ gpg_error_t cmd_readcert (assuan_context_t ctx, char *line)
 	pkcs11h_certificate_t cert = NULL;
 	unsigned char *blob = NULL;
 	size_t blob_size;
+	const char *l;
+
+	l = strgetopt_getopt(line, NULL);
 
 	if (
 		(error = _get_certificate_by_name (
 			ctx,
-			line,
+			l,
 			0,
 			&cert_id,
 			NULL
@@ -852,17 +867,32 @@ gpg_error_t cmd_readkey (assuan_context_t ctx, char *line)
 	gcry_sexp_t sexp = NULL;
 	unsigned char *blob = NULL;
 	size_t blob_size;
+	char *key_hexgrip = NULL;
+	char *keypairinfo = NULL;
+	const char *l;
+	int info = 0;
+	int info_only = 0;
+	char *ser = NULL;
+	size_t ser_len;
+	const struct strgetopt_option options[] = {
+		{"info", strgtopt_no_argument, NULL, &info},
+		{"info-only", strgtopt_no_argument, NULL, &info_only},
+		{NULL, 0, NULL, NULL}
+	};
+
+	l = strgetopt_getopt(line, options);
 
 	if (
 		(error = _get_certificate_by_name (
 			ctx,
-			line,
+			l,
 			0,
 			&cert_id,
 			NULL
 		)) != GPG_ERR_NO_ERROR ||
 		(error = get_cert_sexp (ctx, cert_id, &sexp)) != GPG_ERR_NO_ERROR
 	) {
+		error = GPG_ERR_NOT_FOUND;
 		goto cleanup;
 	}
 
@@ -881,25 +911,94 @@ gpg_error_t cmd_readkey (assuan_context_t ctx, char *line)
 		goto cleanup;
 	}
 
-	if (
-		(error = assuan_send_data(
-			ctx,
-			blob,
-			gcry_sexp_canon_len (blob, 0, NULL, NULL)
-		)) != GPG_ERR_NO_ERROR
-	) {
-		goto cleanup;
+	if (info || info_only) {
+		if (
+			(error = common_map_pkcs11_error (
+				pkcs11h_certificate_serializeCertificateId (
+					NULL,
+					&ser_len,
+					cert_id
+				)
+			)) != GPG_ERR_NO_ERROR
+		) {
+			goto cleanup;
+		}
+
+		if ((ser = (char *)malloc (ser_len)) == NULL) {
+			error = GPG_ERR_ENOMEM;
+			goto cleanup;
+		}
+
+		if (
+			(error = common_map_pkcs11_error (
+				pkcs11h_certificate_serializeCertificateId (
+					ser,
+					&ser_len,
+					cert_id
+				)
+			)) != GPG_ERR_NO_ERROR
+		) {
+			goto cleanup;
+		}
+		if (
+			(key_hexgrip = keyutil_get_cert_hexgrip (sexp)) == NULL ||
+			(keypairinfo = strdup (key_hexgrip)) == NULL ||
+			!encoding_strappend (&keypairinfo, " ") ||
+			!encoding_strappend (&keypairinfo, ser)
+		) {
+			error = GPG_ERR_ENOMEM;
+			goto cleanup;
+		}
+
+		if (
+			(error = assuan_write_status (
+				ctx,
+				"KEYPAIRINFO",
+				keypairinfo
+			)) != GPG_ERR_NO_ERROR
+		) {
+			goto cleanup;
+		}
+	}
+
+	if (!info_only) {
+		if (
+			(error = assuan_send_data(
+				ctx,
+				blob,
+				gcry_sexp_canon_len (blob, 0, NULL, NULL)
+			)) != GPG_ERR_NO_ERROR
+		) {
+			goto cleanup;
+		}
 	}
 
 	error = GPG_ERR_NO_ERROR;
 
 cleanup:
 
+	strgetopt_free(options);
+
 	if (sexp != NULL) {
 		gcry_sexp_release(sexp);
 		sexp = NULL;
 	}
 
+	if (key_hexgrip == NULL) {
+		free(key_hexgrip);
+		key_hexgrip = NULL;
+	}
+
+	if (keypairinfo == NULL) {
+		free(keypairinfo);
+		keypairinfo = NULL;
+	}
+
+	if (ser != NULL) {
+		free(ser);
+		ser = NULL;
+	}
+
 	if (cert_id != NULL) {
 		pkcs11h_certificate_freeCertificateId (cert_id);
 		cert_id = NULL;
@@ -921,32 +1020,20 @@ gpg_error_t cmd_setdata (assuan_context_t ctx, char *line)
 	int append = 0;
 	int index;
 	size_t len;
+	const char *l;
 
-	while (*line != '\x0' && (isspace (*line) || *line == '-')) {
-		if (*line == '-') {
-			static const char *appendprm = "--append ";
-			char *p = line;
+	const struct strgetopt_option options[] = {
+		{"append", strgtopt_no_argument, NULL, &append},
+		{NULL, 0, NULL, NULL}
+	};
 
-			while (*line != '\x0' && !isspace (*line)) {
-				line++;
-			}
-			line++;
-
-			if (!strncmp (p, appendprm, strlen (appendprm))) {
-				p += strlen (appendprm);
-				append = 1;
-			}
-		}
-		else {
-			line++;
-		}
-	}
+	l = strgetopt_getopt(line, options);
 
 	if (!append) {
 		cmd_free_data (ctx);
 	}
 
-	if (!encoding_hex2bin(line, NULL, &len)) {
+	if (!encoding_hex2bin(l, NULL, &len)) {
 		error = GPG_ERR_INV_DATA;
 		goto cleanup;
 	}
@@ -962,7 +1049,7 @@ gpg_error_t cmd_setdata (assuan_context_t ctx, char *line)
 		data->data = (unsigned char *)realloc(data->data, data->size);
 	}
 
-	if (!encoding_hex2bin (line, data->data + index, NULL)) {
+	if (!encoding_hex2bin (l, data->data + index, NULL)) {
 		error = GPG_ERR_INV_DATA;
 		goto cleanup;
 	}
@@ -971,11 +1058,13 @@ gpg_error_t cmd_setdata (assuan_context_t ctx, char *line)
 
 cleanup:
 
+	strgetopt_free(options);
+
 	return gpg_error (error);
 }
 
-/** Sign data (set by SETDATA) with certificate id in line. */
-gpg_error_t cmd_pksign (assuan_context_t ctx, char *line)
+static
+gpg_error_t _cmd_pksign_type (assuan_context_t ctx, char *line, int typehint)
 {
 	static const unsigned char rmd160_prefix[] = /* (1.3.36.3.2.1) */
 		{ 0x30, 0x21, 0x30, 0x09, 0x06, 0x05, 0x2b, 0x24, 0x03,
@@ -1012,7 +1101,6 @@ gpg_error_t cmd_pksign (assuan_context_t ctx, char *line)
 	int session_locked = 0;
 	unsigned char *sig = NULL;
 	size_t sig_len;
-	char hash[100] = "";
 	enum {
 		INJECT_NONE,
 		INJECT_RMD160,
@@ -1023,41 +1111,28 @@ gpg_error_t cmd_pksign (assuan_context_t ctx, char *line)
 		INJECT_SHA384,
 		INJECT_SHA512
 	} inject = INJECT_NONE;
+	char *hash = NULL;
+	const char *l;
+	const struct strgetopt_option options[] = {
+		{"hash", strgtopt_required_argument, &hash, NULL},
+		{NULL, 0, NULL, NULL}
+	};
 
 	if (data->data == NULL) {
 		error = GPG_ERR_INV_DATA;
 		goto cleanup;
 	}
 
-	while (*line != '\x0' && (isspace (*line) || *line == '-')) {
-		if (*line == '-') {
-			static const char *hashprm = "--hash=";
-			char *p = line;
+	l = strgetopt_getopt(line, options);
 
-			while (*line != '\x0' && !isspace (*line)) {
-				line++;
-			}
-			line++;
-
-			if (!strncmp (p, hashprm, strlen (hashprm))) {
-				p += strlen (hashprm);
-				*(line-1) = '\0';
-				snprintf (hash, sizeof(hash), "%s", p);
-			}
-		}
-		else {
-			line++;
-		}
-	}
-
-	if (*line == '\x0') {
+	if (*l == '\x0') {
 		error = GPG_ERR_INV_DATA;
 		goto cleanup;
 	}
 	/*
 	 * sender prefixed data with algorithm OID
 	 */
-	if (strcmp(hash, "")) {
+	if (hash != NULL) {
 		if (!strcmp(hash, "rmd160") && data->size == (0x14 + sizeof(rmd160_prefix)) &&
 			!memcmp (data->data, rmd160_prefix, sizeof (rmd160_prefix))) {
 			inject = INJECT_NONE;
@@ -1134,6 +1209,13 @@ gpg_error_t cmd_pksign (assuan_context_t ctx, char *line)
 			 * gnupg's scdaemon forces to SHA1
 			 */
 			inject = INJECT_SHA1;
+
+			/* When doing auth operation, hash algorithm prefix detection does not work
+			 * but data always comes with algorithm appended, so do not append anything
+			 * by default. */
+			if (typehint == OPENPGP_AUTH) {
+				inject = INJECT_NONE;
+			}
 		}
 	}
 
@@ -1196,8 +1278,8 @@ gpg_error_t cmd_pksign (assuan_context_t ctx, char *line)
 	if (
 		(error = _get_certificate_by_name (
 			ctx,
-			line,
-			OPENPGP_SIGN,
+			l,
+			typehint,
 			&cert_id,
 			NULL
 		)) != GPG_ERR_NO_ERROR
@@ -1295,9 +1377,23 @@ cleanup:
 		_data = NULL;
 	}
 
+	strgetopt_free(options);
+
 	return gpg_error (error);
 }
 
+/** Sign data (set by SETDATA) with certificate id in line. */
+gpg_error_t cmd_pksign (assuan_context_t ctx, char *line)
+{
+	return _cmd_pksign_type(ctx, line, OPENPGP_SIGN);
+}
+
+/** Sign data (set by SETDATA) with certificate id in line. */
+gpg_error_t cmd_pkauth (assuan_context_t ctx, char *line)
+{
+	return _cmd_pksign_type(ctx, line, OPENPGP_AUTH);
+}
+
 /** Decrypt data (set by SETDATA) with certificate id in line. */
 gpg_error_t cmd_pkdecrypt (assuan_context_t ctx, char *line)
 {
@@ -1309,6 +1405,9 @@ gpg_error_t cmd_pkdecrypt (assuan_context_t ctx, char *line)
 	int session_locked = 0;
 	cmd_data_t *data = (cmd_data_t *)assuan_get_pointer (ctx);
 	cmd_data_t _data;
+	const char *l;
+
+	l = strgetopt_getopt(line, NULL);
 
 	if (
 		data == NULL ||
@@ -1340,7 +1439,7 @@ gpg_error_t cmd_pkdecrypt (assuan_context_t ctx, char *line)
 	if (
 		(error = _get_certificate_by_name (
 			ctx,
-			line,
+			l,
 			OPENPGP_ENCR,
 			&cert_id,
 			NULL
@@ -1461,17 +1560,20 @@ gpg_error_t cmd_getinfo (assuan_context_t ctx, char *line)
 {
 	cmd_data_t *data = (cmd_data_t *)assuan_get_pointer (ctx);
 	gpg_err_code_t error = GPG_ERR_GENERAL;
+	const char *l;
 
-	if (!strcmp (line, "version")) {
+	l = strgetopt_getopt(line, NULL);
+
+	if (!strcmp (l, "version")) {
 		char *s = PACKAGE_VERSION;
 		error = assuan_send_data(ctx, s, strlen (s));
 	}
-	else if (!strcmp (line, "pid")) {
+	else if (!strcmp (l, "pid")) {
 		char buf[50];
 		snprintf (buf, sizeof (buf), "%lu", (unsigned long)getpid());
 		error = assuan_send_data(ctx, buf, strlen (buf));
 	}
-	else if (!strcmp (line, "socket_name")) {
+	else if (!strcmp (l, "socket_name")) {
 		const char *s = data->socket_name;
 
 		if (s == NULL) {
@@ -1481,7 +1583,7 @@ gpg_error_t cmd_getinfo (assuan_context_t ctx, char *line)
 			error = assuan_send_data(ctx, s, strlen (s));
 		}
 	}
-	else if (!strcmp (line, "status")) {
+	else if (!strcmp (l, "status")) {
 		pkcs11h_certificate_id_list_t user_certificates = NULL;
 		char flag = 'r';
 
@@ -1506,7 +1608,7 @@ gpg_error_t cmd_getinfo (assuan_context_t ctx, char *line)
 
 		error = assuan_send_data(ctx, &flag, 1);
 	}
-	else if (!strcmp (line, "reader_list")) {
+	else if (!strcmp (l, "reader_list")) {
 		error = GPG_ERR_NO_DATA;
 	}
 	else {
@@ -1516,6 +1618,190 @@ gpg_error_t cmd_getinfo (assuan_context_t ctx, char *line)
 	return gpg_error (error);
 }
 
+gpg_error_t cmd_keyinfo (assuan_context_t ctx, char *line)
+{
+	gpg_err_code_t error = GPG_ERR_GENERAL;
+	pkcs11h_certificate_id_list_t user_certificates = NULL;
+	pkcs11h_certificate_id_list_t curr_cert;
+	char *list = NULL;
+	int data_arg = 0;
+	const char *l;
+	char *serial;
+	int found;
+	int filter;
+
+	const struct strgetopt_option options[] = {
+		{"list", strgtopt_optional_argument, &list, NULL},
+		{"data", strgtopt_no_argument, NULL, &data_arg},
+		{NULL, 0, NULL, NULL}
+	};
+
+	l = strgetopt_getopt(line, options);
+
+	if (list == NULL) {
+	}
+	else if (!strcmp(list, "")) {
+		filter = 0;
+	}
+	else if (!strcmp(list, "auth")) {
+		filter = OPENPGP_AUTH;
+	}
+	else if (!strcmp(list, "encr")) {
+		filter = OPENPGP_ENCR;
+	}
+	else if (!strcmp(list, "sign")) {
+		filter = OPENPGP_SIGN;
+	}
+	else {
+		goto cleanup;
+	}
+
+	found = list != NULL;
+
+	if (
+		(error = common_map_pkcs11_error (
+			pkcs11h_certificate_enumCertificateIds (
+				PKCS11H_ENUM_METHOD_CACHE_EXIST,
+				ctx,
+				PKCS11H_PROMPT_MASK_ALLOW_ALL,
+				NULL,
+				&user_certificates
+			)
+		)) != GPG_ERR_NO_ERROR
+	) {
+		goto cleanup;
+	}
+
+	for (
+		curr_cert = user_certificates;
+		curr_cert != NULL;
+		curr_cert = curr_cert->next
+	) {
+		char *certid = NULL;
+		char *key_hexgrip = NULL;
+		char *keyinfo_line = NULL;
+		gcry_sexp_t sexp = NULL;
+		size_t ser_len;
+		int print = 0;
+
+		if ((error = get_cert_sexp (ctx, curr_cert->certificate_id, &sexp)) != GPG_ERR_NO_ERROR) {
+			goto retry;
+		}
+
+		if ((key_hexgrip = keyutil_get_cert_hexgrip (sexp)) == NULL) {
+			error = GPG_ERR_ENOMEM;
+			goto retry;
+		}
+
+		if (
+			(error = common_map_pkcs11_error (
+				pkcs11h_certificate_serializeCertificateId (
+					NULL,
+					&ser_len,
+					curr_cert->certificate_id
+				)
+			)) != GPG_ERR_NO_ERROR
+		) {
+			goto retry;
+		}
+
+		if ((certid = (char *)malloc (ser_len)) == NULL	) {
+			error = GPG_ERR_ENOMEM;
+			goto retry;
+		}
+
+		if (
+			(error = common_map_pkcs11_error (
+				pkcs11h_certificate_serializeCertificateId (
+					certid,
+					&ser_len,
+					curr_cert->certificate_id
+				)
+			)) != GPG_ERR_NO_ERROR
+		) {
+			goto retry;
+		}
+
+		if ((error = get_serial_of_tokenid(curr_cert->certificate_id->token_id, &serial)) != GPG_ERR_NO_ERROR) {
+			goto cleanup;
+		}
+
+		if (list == NULL) {
+			if (!strcmp(l, key_hexgrip)) {
+				found=1;
+				print=1;
+			}
+		}
+		else {
+			print=1;
+		}
+
+		if (data_arg &&  print) {
+			if (
+				(keyinfo_line = strdup (key_hexgrip)) == NULL ||
+				!encoding_strappend (&keyinfo_line, " T ") ||
+				!encoding_strappend (&keyinfo_line, serial) ||
+				!encoding_strappend (&keyinfo_line, " ") ||
+				!encoding_strappend (&keyinfo_line, certid)
+			) {
+				error = GPG_ERR_ENOMEM;
+				goto retry;
+			}
+
+			if (
+				(error = assuan_write_status(
+					ctx,
+					"KEYINFO",
+					keyinfo_line
+				)) != GPG_ERR_NO_ERROR
+			) {
+				goto cleanup;
+			}
+		}
+
+		error = GPG_ERR_NO_ERROR;
+
+	retry:
+
+		if (keyinfo_line != NULL) {
+			free (keyinfo_line);
+			keyinfo_line = NULL;
+		}
+
+		if (certid != NULL) {
+			free (certid);
+			certid = NULL;
+		}
+
+		if (serial != NULL) {
+			free(serial);
+			serial = NULL;
+		}
+
+		if (key_hexgrip != NULL) {
+			free (key_hexgrip);
+			key_hexgrip = NULL;
+		}
+
+		if (error != GPG_ERR_NO_ERROR) {
+			goto cleanup;
+		}
+	}
+
+	error = found ? GPG_ERR_NO_ERROR : GPG_ERR_NOT_FOUND;
+
+cleanup:
+
+	strgetopt_free(options);
+
+	if (user_certificates != NULL) {
+		pkcs11h_certificate_freeCertificateIdList (user_certificates);
+		user_certificates = NULL;
+	}
+
+	return gpg_error (error);
+}
+
 gpg_error_t cmd_restart (assuan_context_t ctx, char *line)
 {
 	(void)ctx;
@@ -1529,8 +1815,11 @@ gpg_error_t cmd_getattr (assuan_context_t ctx, char *line)
 	pkcs11h_certificate_id_list_t user_certificates = NULL;
 	char *serial = NULL;
 	gpg_err_code_t error = GPG_ERR_GENERAL;
+	const char *l;
 
-	if (!strcmp (line, "SERIALNO")) {
+	l = strgetopt_getopt(line, NULL);
+
+	if (!strcmp (l, "SERIALNO")) {
 		if (
 			(error = get_serial(ctx, &serial)) != GPG_ERR_NO_ERROR
 		) {
@@ -1549,7 +1838,7 @@ gpg_error_t cmd_getattr (assuan_context_t ctx, char *line)
 			}
 		}
 	}
-	else if (!strcmp (line, "KEY-FPR")) {
+	else if (!strcmp (l, "KEY-FPR")) {
 		if (
 			(error = common_map_pkcs11_error (
 				pkcs11h_certificate_enumCertificateIds (
@@ -1569,7 +1858,7 @@ gpg_error_t cmd_getattr (assuan_context_t ctx, char *line)
 			goto cleanup;
 		}
 	}
-	else if (!strcmp (line, "CHV-STATUS")) {
+	else if (!strcmp (l, "CHV-STATUS")) {
 		if (
 			(error = assuan_write_status(
 				ctx,
@@ -1580,7 +1869,7 @@ gpg_error_t cmd_getattr (assuan_context_t ctx, char *line)
 			goto cleanup;
 		}
 	}
-	else if (!strcmp (line, "DISP-NAME")) {
+	else if (!strcmp (l, "DISP-NAME")) {
 		if (
 			(error = assuan_write_status(
 				ctx,
@@ -1591,7 +1880,7 @@ gpg_error_t cmd_getattr (assuan_context_t ctx, char *line)
 			goto cleanup;
 		}
 	}
-	else if (!strcmp (line, "KEY-ATTR")) {
+	else if (!strcmp (l, "KEY-ATTR")) {
 		int i;
 		for (i=0;i<3;i++) {
 			char buffer[1024];
@@ -1610,7 +1899,7 @@ gpg_error_t cmd_getattr (assuan_context_t ctx, char *line)
 			}
 		}
 	}
-	else if (!strcmp (line, "EXTCAP")) {
+	else if (!strcmp (l, "EXTCAP")) {
 		int i;
 		for (i=0;i<3;i++) {
 			char buffer[1024];
@@ -1655,8 +1944,11 @@ cleanup:
 gpg_error_t cmd_setattr (assuan_context_t ctx, char *line)
 {
 	gpg_err_code_t error = GPG_ERR_GENERAL;
+	const char *l;
 
-	if (!strncmp (line, "CHV-STATUS-1 ", 13)) {
+	l = strgetopt_getopt(line, NULL);
+
+	if (!strncmp (l, "CHV-STATUS-1 ", 13)) {
 	}
 	else {
 		error = GPG_ERR_INV_DATA;
@@ -1682,44 +1974,34 @@ gpg_error_t cmd_genkey (assuan_context_t ctx, char *line)
 	char *e_resp = strdup ("e ");
 	unsigned char *blob = NULL;
 	char *serial = NULL;
-	char *key = NULL;
+	const char *key = NULL;
 	size_t blob_size;
-	char timestamp[100] = {0};
+	char *timestamp = NULL;
+	char _timestamp[100];
+	const char *l;
 
-	while (*line != '\x0' && !isdigit (*line)) {
-		if (*line == '-') {
-			static const char *ts = "--timestamp=";
-			char *p = line;
+	const struct strgetopt_option options[] = {
+		{"timestamp", strgtopt_required_argument, &timestamp, NULL},
+		{NULL, 0, NULL, NULL}
+	};
 
-			while (*line != '\x0' && !isspace (*line)) {
-				line++;
-			}
-			line++;
+	l = strgetopt_getopt(line, options);
 
-			if (!strncmp (p, ts, strlen (ts))) {
-				p += strlen (ts);
-				sprintf (timestamp, "%d", (int)isotime2epoch (p));
-			}
-		}
-		else {
-			line++;
-		}
-	}
-
-	if (*line == '\x0') {
+	if (*l == '\x0') {
 		error = GPG_ERR_INV_DATA;
 		goto cleanup;
 	}
 
-	if (strlen (timestamp) == 0) {
-		sprintf (timestamp, "%d", (int)time (NULL));
+	if (timestamp == NULL) {
+		sprintf (_timestamp, "%d", (int)time (NULL));
+		timestamp = _timestamp;
 	}
 
 	if (
 		(error = _get_certificate_by_name (
 			ctx,
 			NULL,
-			atoi(line),
+			atoi(l),
 			&cert_id,
 			&key
 		)) != GPG_ERR_NO_ERROR
@@ -1863,6 +2145,8 @@ cleanup:
 		serial = NULL;
 	}
 
+	strgetopt_free(options);
+
 	return gpg_error (error);
 }
 
