@@ -100,7 +100,7 @@ keyutil_get_cert_mpi (
 	}
 
 	if (gnutls_x509_crt_get_pk_rsa_raw (cert, &m, &e) != GNUTLS_E_SUCCESS) {
-		error = GPG_ERR_BAD_KEY;
+		error = GPG_ERR_WRONG_PUBKEY_ALGO;
 		m.data = NULL;
 		e.data = NULL;
 		goto cleanup;
@@ -220,7 +220,7 @@ cleanup:
 }
 /**
    Convert X.509 RSA public key into gcrypt internal sexp form. Only RSA
-   public keys are accepted at the moment. The resul is stored in *sexp,
+   public keys are accepted at the moment. The result is stored in *sexp,
    which must be freed (using ) when not needed anymore. *sexp must be
    NULL on entry, since it is overwritten.
 */
