@@ -45,6 +45,7 @@
 #include <getopt.h>
 #include <errno.h>
 #include <pkcs11-helper-1.0/pkcs11h-core.h>
+#include <pkcs11-helper-1.0/pkcs11h-token.h>
 #if !defined(HAVE_W32_SYSTEM)
 #include <sys/stat.h>
 #include <sys/types.h>
@@ -125,7 +126,7 @@ register_commands (const assuan_context_t ctx)
 		{ "KEY-DATA",	NULL, NULL },
 		{ "SETDATA",	cmd_setdata, NULL },
 		{ "PKSIGN",	cmd_pksign, NULL },
-		{ "PKAUTH",	NULL, NULL },
+		{ "PKAUTH",	cmd_pkauth, NULL },
 		{ "PKDECRYPT",	cmd_pkdecrypt, NULL },
 		{ "INPUT",	NULL, NULL }, 
 		{ "OUTPUT",	NULL, NULL }, 
@@ -140,6 +141,7 @@ register_commands (const assuan_context_t ctx)
 		{ "LOCK",	NULL, NULL },
 		{ "UNLOCK",	NULL, NULL },
 		{ "GETINFO",	cmd_getinfo, NULL },
+		{ "KEYINFO",	cmd_keyinfo, NULL },	/* gnupg-2.3.x */
 		{ "RESTART",	cmd_restart, NULL },
 		{ "DISCONNECT",	cmd_null, NULL },
 		{ "APDU",	NULL, NULL },
@@ -575,15 +577,46 @@ pkcs11_token_prompt_hook (
 	assuan_context_t ctx = user_data;
 	int rc;
 	int ret = FALSE;
+	char *ser = NULL;
+	size_t n;
 
 	(void)global_data;
 	(void)retry;
 
+	if (
+		(rc = common_map_pkcs11_error(
+			pkcs11h_token_serializeTokenId(
+				NULL,
+				&n,
+				token
+			)
+		)) != GPG_ERR_NO_ERROR
+	) {
+		goto cleanup;
+	}
+
+	if ((ser = (char *)malloc(n)) == NULL) {
+		rc = GPG_ERR_ENOMEM;
+		goto cleanup;
+	}
+
+	if (
+		(rc = common_map_pkcs11_error(
+			pkcs11h_token_serializeTokenId(
+				ser,
+				&n,
+				token
+			)
+		)) != GPG_ERR_NO_ERROR
+	) {
+		goto cleanup;
+	}
+
 	snprintf (
 		cmd,
 		sizeof(cmd),
-		"NEEDPIN Please insert token '%s' !!!DO NOT ENTER PIN HERE!!!!",
-		token->display
+		"NEEDPIN %s",
+		ser
 	);
 
 	if ((rc = assuan_inquire (ctx, cmd, &user_read, &user_read_len, 1024))) {
@@ -591,7 +624,7 @@ pkcs11_token_prompt_hook (
 		goto cleanup;
 	}
 
-	if (!strcmp ((char *)user_read, "cancel")) {
+	if (!strcmp ((char *)user_read, "cancel") || !strcmp((char *)user_read, "no")) {
 		goto cleanup;
 	}
 
@@ -599,6 +632,11 @@ pkcs11_token_prompt_hook (
 
 cleanup:
 
+	if (ser != NULL) {
+		free(ser);
+		ser = NULL;
+	}
+
 	if (user_read != NULL) {
 		memset (user_read, 0, strlen ((char *)user_read));
 		free (user_read);
@@ -618,29 +656,79 @@ pkcs11_pin_prompt_hook (
 	char * const pin,
 	const size_t max_pin
 ) {
+	global_t *global = (global_t *)global_data;
 	char cmd[1024];
 	assuan_context_t ctx = user_data;
 	unsigned char *pin_read = NULL;
 	size_t pin_len;
 	int rc;
 	int ret = FALSE;
+	char *ser = NULL;
+	size_t n;
 
-	(void)global_data;
-
-	snprintf (
-		cmd,
-		sizeof(cmd),
-		"NEEDPIN PIN required for token '%s' (try %u)",
-		token->display,
-		retry
-	);
-
-	if ((rc = assuan_inquire (ctx, cmd, &pin_read, &pin_len, 1024))) {
-		common_log (LOG_WARNING,"PIN inquire error: %d", rc);
+	if (
+		(rc = common_map_pkcs11_error(
+			pkcs11h_token_serializeTokenId(
+				NULL,
+				&n,
+				token
+			)
+		)) != GPG_ERR_NO_ERROR
+	) {
 		goto cleanup;
 	}
 
+	if ((ser = (char *)malloc(n)) == NULL) {
+		rc = GPG_ERR_ENOMEM;
+		goto cleanup;
+	}
+
+	if (
+		(rc = common_map_pkcs11_error(
+			pkcs11h_token_serializeTokenId(
+				ser,
+				&n,
+				token
+			)
+		)) != GPG_ERR_NO_ERROR
+	) {
+		goto cleanup;
+	}
+
+	if (!global->config.use_gnupg_pin_cache) {
+		common_log (LOG_WARNING, "PIN cache disabled");
+		pin_len = 0;
+	}
+	else {
+		snprintf (
+			cmd,
+			sizeof(cmd),
+			"PINCACHE_GET %s",
+			ser
+		);
+
+		if ((rc = assuan_inquire (ctx, cmd, &pin_read, &pin_len, 1024))) {
+			common_log (LOG_WARNING,"PIN cache inquire error: %d [ignored]", rc);
+			pin_len = 0;
+		}
+	}
+
 	if (pin_len==0 || (pin_len+1 > max_pin)) {
+		snprintf (
+			cmd,
+			sizeof(cmd),
+			"NEEDPIN %s",
+			ser
+		);
+
+		if ((rc = assuan_inquire (ctx, cmd, &pin_read, &pin_len, 1024))) {
+			common_log (LOG_WARNING,"PIN inquire error: %d", rc);
+			goto cleanup;
+		}
+	}
+
+	if (pin_len==0 || (pin_len+1 > max_pin)) {
+		rc = GPG_ERR_TOO_LARGE;
 		goto cleanup;
 	}
 
@@ -650,6 +738,133 @@ pkcs11_pin_prompt_hook (
 
 cleanup:
 
+	if (ser != NULL) {
+		free(ser);
+		ser = NULL;
+	}
+
+	if (pin_read != NULL) {
+		memset (pin_read, 0, strlen ((char *)pin_read));
+		free (pin_read);
+		pin_read = NULL;
+	}
+
+	return ret;
+}
+
+static
+PKCS11H_BOOL
+pkcs11_key_prompt_hook (
+	void * const global_data,
+	void * const user_data,
+	const pkcs11h_token_id_t token,
+	const char * const label,
+	const unsigned retry,
+	char * const pin,
+	const size_t max_pin
+) {
+	global_t *global = (global_t *)global_data;
+	char keyid[1024];
+	char cmd[1024];
+	assuan_context_t ctx = user_data;
+	unsigned char *pin_read = NULL;
+	size_t pin_len;
+	int rc;
+	int ret = FALSE;
+	char *ser = NULL;
+	size_t n;
+
+	snprintf(keyid, sizeof(keyid), "%s", label);
+	keyid[sizeof(keyid)-1] = '\0';
+
+	{
+		size_t i;
+		for (i=0;i<strlen(keyid); i++) {
+			if (isspace(keyid[i])) {
+				keyid[i] = '_';
+			}
+		}
+	}
+
+	if (
+		(rc = common_map_pkcs11_error(
+			pkcs11h_token_serializeTokenId(
+				NULL,
+				&n,
+				token
+			)
+		)) != GPG_ERR_NO_ERROR
+	) {
+		goto cleanup;
+	}
+
+	if ((ser = (char *)malloc(n)) == NULL) {
+		rc = GPG_ERR_ENOMEM;
+		goto cleanup;
+	}
+
+	if (
+		(rc = common_map_pkcs11_error(
+			pkcs11h_token_serializeTokenId(
+				ser,
+				&n,
+				token
+			)
+		)) != GPG_ERR_NO_ERROR
+	) {
+		goto cleanup;
+	}
+
+	if (!global->config.use_gnupg_pin_cache) {
+		common_log (LOG_WARNING, "PIN cache disabled");
+		pin_len = 0;
+	}
+	else {
+		snprintf (
+			cmd,
+			sizeof(cmd),
+			"PINCACHE_GET %s/%s",
+			ser,
+			keyid
+		);
+
+		if ((rc = assuan_inquire (ctx, cmd, &pin_read, &pin_len, 1024))) {
+			common_log (LOG_WARNING,"PIN cache inquire error: %d [ignored]", rc);
+			pin_len = 0;
+		}
+	}
+
+	if (pin_len==0 || (pin_len+1 > max_pin)) {
+		snprintf (
+			cmd,
+			sizeof(cmd),
+			"NEEDPIN %s/%s",
+			ser,
+			keyid
+		);
+
+		if ((rc = assuan_inquire (ctx, cmd, &pin_read, &pin_len, 1024))) {
+			common_log (LOG_WARNING,"PIN inquire error: %d", rc);
+			goto cleanup;
+		}
+	}
+
+	if (pin_len==0 || (pin_len+1 > max_pin)) {
+		rc = GPG_ERR_TOO_LARGE;
+		goto cleanup;
+	}
+
+	strcpy (pin, (char *)pin_read);
+
+	ret = TRUE;
+
+cleanup:
+
+	if (ser != NULL) {
+		free(ser);
+		ser = NULL;
+	}
+
 	if (pin_read != NULL) {
 		memset (pin_read, 0, strlen ((char *)pin_read));
 		free (pin_read);
@@ -1118,8 +1333,17 @@ int main (int argc, char *argv[])
 
 	pkcs11h_setLogLevel (global.config.verbose ? PKCS11H_LOG_DEBUG2 : PKCS11H_LOG_INFO);
 	pkcs11h_setLogHook (pkcs11_log_hook, NULL);
-	pkcs11h_setTokenPromptHook (pkcs11_token_prompt_hook, NULL);
-	pkcs11h_setPINPromptHook (pkcs11_pin_prompt_hook, NULL);
+	pkcs11h_setTokenPromptHook (pkcs11_token_prompt_hook, &global);
+	pkcs11h_setPINPromptHook (pkcs11_pin_prompt_hook, &global);
+#ifdef PKCS11H_PROPERTY_KEY_PROMPT_HOOK
+	{
+		void *p;
+		p = &pkcs11_key_prompt_hook;
+		pkcs11h_setProperty(PKCS11H_PROPERTY_KEY_PROMPT_HOOK, &p, sizeof(p));
+		p = &global;
+		pkcs11h_setProperty(PKCS11H_PROPERTY_KEY_PROMPT_HOOK_DATA, &p, sizeof(p));
+	}
+#endif
 	pkcs11h_setProtectedAuthentication (TRUE);
 	pkcs11h_setPINCachePeriod(global.config.pin_cache);
 
